#
##################################################################
#
# File: Psad.pm
#
# Purpose: This is the Psad.pm perl module.  It contains functions
#          that are reused by the various psad daemons.
#
# Author: Michael Rash (mbr@cipherdyne.org)
#
# Version: 1.2.3
#
##################################################################
#
# $Id: Psad.pm,v 1.71 2003/09/13 01:36:53 mbr Exp $
#

package Psad;

use lib '/usr/lib/psad';
use Exporter;
use Text::Wrap;
use Unix::Syslog qw(:subs :macros);
use Carp;
use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
    buildconf
    defined_vars
    writepid
    writecmdline
    unique_pid
    warn_handler
    die_handler
    check_commands
    check_firewall_rules
    sendmail
    psyslog
    logr
);

$VERSION = '1.2.3';

### define the main psad logging directory;
### this is used by the warn() and die() handlers
my $PSAD_DIR = '/var/log/psad';

### used by logr()
my $SUB_TAB = '    ';

### subroutines ###
sub buildconf() {
    my ($config_hr, $cmds_hr, $conf_file) = @_;
    open C, "< $conf_file" or croak "  ** Could not open " .
        "config file $conf_file: $!";
    my @lines = <C>;
    close C;
    for my $line (@lines) {
        chomp $line;
        next if ($line =~ /^\s*#/);
        if ($line =~ /^(\S+)\s+(.*?)\;/) {
            my $varname = $1;
            my $val     = $2;
            if ($val =~ m|/.+| && $varname =~ /^(\w+)Cmd$/) {
                ### found a command
                $cmds_hr->{$1} = $val;
            } else {
                $config_hr->{$varname} = $val;
            }
        }
    }
    return;
}

### check to make sure all required varables are defined in the config
### this subroutine is passed different variables by each script that
### correspond to only those variables needed be each script).
sub defined_vars() {
    my ($config_hr, $conf_file, $varnames_aref) = @_;
    for my $var (@$varnames_aref) {
        unless (defined $config_hr->{$var}) {  ### missing var
            croak " ** The config file \"$conf_file\" does not " .
                  "contain the\nvariable: \"$var\".  Exiting!";
        }
    }
    return;
}

### check paths to commands and attempt to correct if any are wrong.
sub check_commands() {
    my $cmds_href = shift;
    my $caller = $0;
    my @path = qw(
        /bin
        /sbin
        /usr/bin
        /usr/sbin
        /usr/local/bin
        /usr/local/sbin
    );
    CMD: for my $cmd (keys %$cmds_href) {
        unless (-x $cmds_href->{$cmd}) {
            my $found = 0;
            PATH: for my $dir (@path) {
                if (-x "${dir}/${cmd}") {
                    $cmds_href->{$cmd} = "${dir}/${cmd}";
                    $found = 1;
                    last PATH;
                }
            }
            unless ($found) {
                next CMD if $cmd eq 'whois_psad';
                croak " **  ($caller): Could not find $cmd ",
                    "anywhere!!!\n    Please edit the config section ",
                     "to include the path to $cmd.";
            }
        }
        unless (-x $cmds_href->{$cmd}) {
            croak " **  ($caller):  $cmd is located at ",
                "$cmds_href->{$cmd}, but is not executable\n",
                "    by uid: $<";
        }
    }
    return;
}

sub check_firewall_rules() {
    my ($fw_msg_search, $emails_str, $fw_check_file, $cmds_href) = @_;
    my @localips;
    ### see which of the two firewall commands we are able
    ### to execute.
    ### return value from system call is a 16 bit number,
    ### the high byte is the real return value
    my $ipchains_rv = (system "$cmds_href->{'ipchains'} -nL " .
        "> /dev/null 2>&1") >> 8;

    my @localips_tmp = `$cmds_href->{'ifconfig'} -a`;

    for my $line (@localips_tmp) {
        chomp $line;
        if ($line =~ /inet\s+addr:(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/) {
            push @localips, $1;
        }
    }

    if (! $ipchains_rv) {
        if (&check_ipchains_rules(
                \@localips,
                $fw_check_file,
                $cmds_href)) {
            return 1;
        } else {
            &sendmail(" ** psad: firewall ruleset warning!", $fw_check_file,
                $emails_str, $cmds_href->{'mail'});
            return 0;
        }
    } else {
        print localtime() . " ** Could not run ipchains or " .
            "iptables command!\n";
        return 0;
    }
    return 1;
}

sub check_ipchains_rules() {
    my ($localips_aref, $fw_check_file, $cmds_href) = @_;
    #  target     prot opt  source        destination     ports
    # DROP    tcp  ----l-   anywhere      anywhere        any ->   telnet

    #Chain input (policy ACCEPT):
    #target     prot opt     source       destination    ports
    #ACCEPT     tcp  ------  0.0.0.0/0    0.0.0.0/0      * ->   22
    #DROP       tcp  ----l-  0.0.0.0/0    0.0.0.0/0      * ->   *
    my @rules = `$cmds_href->{'ipchains'} -nL`;
    FWPARSE: for my $rule (@rules) {
        chomp $rule;
        next FWPARSE if ($rule =~ /^Chain/ || $rule =~ /^target/);
        if ($rule =~ /^(\w+)\s+(\w+)\s+(\S+)\s+\S+\s+(\S+)\s+(\*)\s+\-\>\s+(\*)/) {
            my ($target, $proto, $opt, $dst, $srcpt, $dstpt) =
                ($1, $2, $3, $4, $5, $6);
            if ($target =~ /drop|reject|deny/i && $proto =~ /all|tcp/
                    && $opt =~ /....l./) {
                if (&check_destination($dst, $localips_aref)) {
                    &logr(" .. Your firewall ruleset looks good.  Unauthorized tcp" .
                        " packets will be dropped and logged.\n",
                        {$fw_check_file => 0});
                    return 1;
                }
            }
        } elsif ($rule =~ /^(\w+)\s+(\w+)\s+(\S+)\s+\S+\s+(\S+)\s+(n\/a)/) {
            ### kernel 2.2.14 (and others) show "n/a" instead of "*"
            my ($target, $proto, $opt, $dst, $ports) =
                ($1, $2, $3, $4, $5);
            if ($target =~ /drop|reject|deny/i && $proto =~ /all|tcp/
                        && $opt =~ /....l./) {
                if (&check_destination($dst, $localips_aref)) {
                    &logr(" .. Your firewall ruleset looks good.  Unauthorized tcp" .
                        " packets will be dropped and logged.\n",
                        {$fw_check_file => 0});
                    return 1;
                }
            }
        }
    }
    &logr(" .. Your firewall does not include rules that will log " .
        "dropped/rejected packets.  Psad will not work with such a " .
        "firewall ruleset.\n",
        {$fw_check_file => 0});
    return 0;
}

### this needs to be re-worked
sub check_destination() {
    my ($dst, $localips_aref) = @_;
    return 1 if ($dst =~ /0\.0\.0\.0\/0/);
    for my $ip (@$localips_aref) {
        my ($oct) = ($ip =~ /^(\d{1,3}\.\d{1,3})/);
        return 1 if ($dst =~ /^$oct/);
    }
    return 0;
}

### make sure pid is unique
sub unique_pid() {
    my $pidfile = shift;
    if (-e $pidfile) {
        my $caller = $0;
        open PIDFILE, "< $pidfile";
        my $pid = <PIDFILE>;
        close PIDFILE;
        chomp $pid;
        if (kill 0, $pid) {  # psad is already running
            croak "\n **  $caller (pid: $pid) is already " .
                  "running!  Exiting.\n";
        }
    }
    return;
}

### write the pid to the pid file
sub writepid() {
    my $pidfile = shift;
    my $caller = $0;
    open PIDFILE, "> $pidfile" or croak " **  $caller: Could not " .
        "open pidfile $pidfile: $!\n";
    print PIDFILE $$ . "\n";
    close PIDFILE;
    chmod 0600, $pidfile;
    return;
}

### write command line to cmd file
sub writecmdline() {
    my ($args_aref, $cmdline_file) = @_;
    open CMD, "> $cmdline_file";
    print CMD "@$args_aref\n";
    close CMD;
    chmod 0600, $cmdline_file;
    return;
}

### write all die messages to a logfile
sub die_handler() {
    my $msg = shift;
    my $caller = $0;
    $caller =~ s|^.*/||;
    open D, ">> ${PSAD_DIR}/errs/${caller}.die";
    print D localtime() . " .. $caller (pid: $$): $msg";
    close D;
    die $!;
}

### write all warnings to a logfile
sub warn_handler() {
    my $msg = shift;
    my $caller = $0;
    $caller =~ s|^.*/||;
    open W, ">> ${PSAD_DIR}/errs/${caller}.warn";
    print W localtime() . " .. $caller (pid: $$): $msg";
    close W;
    return;
}

### send mail message to all addresses contained in the
### EMAIL_ADDRESSES variable within psad.conf ($addr_str).
### TODO:  Would it be better to use Net::SMTP here?  Does
### Net::SMTP give you mail queueing like sendmail?
sub sendmail() {
    my ($subject, $body_file, $addr_str, $mailCmd) = @_;
    $body_file = '/dev/null' unless $body_file;
    system "$mailCmd -s \"$subject\" $addr_str < $body_file > /dev/null 2>&1";
    return;
}

### write a message to syslog
sub psyslog() {
    my ($ident, $msg) = @_;
    ### write a message to syslog and return
    openlog $ident, LOG_DAEMON, LOG_LOCAL7;
    syslog LOG_INFO, $msg;
    closelog();
    return;
}

### logging subroutine that handles multiple filehandles
sub logr() {
    my ($msg, $files_href) = @_;
    my $term_width = 70;
    for my $file (keys %$files_href) {
        if ($file eq *STDOUT) {
            if (length($msg) > $term_width) {
                print STDOUT wrap('', $SUB_TAB, $msg);
            } else {
                print STDOUT $msg;
            }
        } elsif ($file eq *STDERR) {
            if (length($msg) > $term_width) {
                print STDERR wrap('', $SUB_TAB, $msg);
            } else {
                print STDERR $msg;
            }
        } else {
            if ($files_href->{$file}) {  ### we are appending to the file
                open F, ">> $file" or croak " ** Could not open file: " .
                    "$file: $!";
            } else {
                open F, "> $file" or croak " ** Could not open file: " .
                    "$file: $!";
            }
            if (length($msg) > $term_width) {
                print F wrap('', $SUB_TAB, $msg);
            } else {
                print F $msg;
            }
            close F;
        }
    }
    return;
}

1;
__END__

=head1 NAME

Psad.pm - Perl extension for the psad (Port Scan Attack Detector) daemons

=head1 SYNOPSIS

  use Psad;
  writepid()
  writecmdline()
  unique_pid()
  psyslog()
  logr()
  check_commands()
  check_firewall_rules()
  $SIG{'__WARN__'} = \&Psad::warn_handler;
  $SIG{'__DIE__'} = \&Psad::die_handler;

=head1 DESCRIPTION

The Psad.pm module contains several subroutines that are used by Port Scan
Attack Detector (psad) daemons.
writepid()  writes process ids to the pid files (e.g. "/var/run/psad.pid").
writecmdline()  writes the psad command line contained within @ARGV[] to the
file "/var/run/psad.cmd".
unique_pid()  makes sure that no other daemon process is already running in
order to guarantee pid uniqueness.
psyslog() an interface to sending messages via syslog.
logr()  a logging subroutine that handles printing to multiple file handles.
check_commands()  check paths to commands and warns if unable to find any 
particular command.
check_firewall_rules()  checks the ipchains/iptables firewall ruleset to see 
if the ruleset is compatible with psad (i.e., if it has been configured to
log and drop all packets that are not explicitly allowed through).  The two
signal handlers warn_handler and die_handler write warnings and die messages
to logfiles in /var/log/psad.

=head1 AUTHOR

Michael Rash, mbr@cipherdyne.org

=head1 SEE ALSO

psad(8).

